"""
Database utility functions.
Copyright (C) 2022 Humankind Investments
"""

import psycopg2
from psycopg2.extras import execute_batch
from psycopg2 import sql

import pandas as pd


def get_data(db_name, db_cred, query):
    """
    Get data from database for given query.

    Parameters
    ----------
    db_name : str
        Name of database on instance.
    db_cred : dict
        Dictionary of database instance credentials, where keys are 
        'endpoint', 'port', 'username', and 'pw'.
    query : str
        Query with which to read data from database.

    Returns
    -------
    pd.DataFrame
        Dataframe of data based on query.
    """

    cursor, conn = database_connect(db_name, db_cred)
    df = pd.read_sql(query, conn)
    cursor.close()
    conn.close()

    return df


def database_connect(db_name, db_cred):
    """
    Connect to database with given name and instance credentials.

    Parameters
    ----------
    db_name : str
        Name of database on instance.
    db_cred : dict
        Dictionary of credentials for database instance, where keys are 
        'endpoint', 'port', 'user', and 'pw', e.g.:
            db_cred = {'endpoint' : [instance-endpoint],
                       'port' : [connection-port-number],
                       'user' : [username],
                       'pw' : [password]}.

    Returns
    -------
    psycopg2.connection.cursor
        Database connection cursor object.
    psycopg2.connection
        Database connection object.
    """

    conn = psycopg2.connect(host = db_cred['endpoint'],
                            port = db_cred['port'],
                            database = db_name,
                            user = db_cred['username'],
                            password = db_cred['pw'],
                            )

    return conn.cursor(), conn


def database_write(df, db_name, db_cred, table_query, table_name):
    """
    Write data to table in database.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to insert into table in database.
    db_name : str
        Name of database on instance.
    db_cred : dict
        Dictionary of database instance credentials, where keys are 
        'endpoint', 'port', 'username', and 'pw'.
    table_query : str
        Initial table creation query.
    table_name : str
        Name of table in which to insert data.
    """

    # connect to database
    cursor, conn = database_connect(db_name, db_cred)

    # create initial table
    cursor.execute(table_query)
    conn.commit()

    # write data to table in DB
    tdict = df.to_dict(orient='records')
    query = insert_data(table_name, df.columns)
    execute_batch(cursor, query, tdict)
    conn.commit()

    cursor.close()
    conn.close()

    
def insert_data(table, columns):
    """
    Create insert query.

    Parameters
    ----------
    table : str
        Name of table in which to insert data.
    columns : list of str
        List of column names in table.
    
    Returns
    -------
    insert : sql query
        SQL insert query.
    """

    insert = sql.SQL('INSERT INTO {} ({}) VALUES({});').format(
        sql.Identifier(table),
        sql.SQL(',').join(map(sql.Identifier, columns)),
        sql.SQL(',').join(map(sql.Placeholder, columns))
        )

    return insert


def connect_dynamo_table(session, table_name):
    """
    Connect to DynamoDB table.

    Parameters
    ----------
    session : boto3.session.Session
        Boto3 session object.
    table_name : str
        Name of table.
    
    Returns
    -------
    DynamoDB.Table
        DynamoDB table resource.
    """
    
    # connect to dynamodb table
    dynamodb = session.resource('dynamodb')
    table = dynamodb.Table(table_name)

    return table



"""
Unit tests for date utilities
"""

import unittest
from datetime import datetime, timedelta
from pytz import timezone

import os, sys
root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.join(root, 'src', 'utilities_hki'))

from date_utils import get_next_market_day
from date_utils import get_prev_market_day


eastern = timezone('US/Eastern')


def day_generator(year, month, weekday):
    """
    Generate all dates in a month that fall on a given weekday.
    """
    if isinstance(weekday, str):
        weekday = ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'].index(weekday.lower())

    d = datetime(year, month, 1)
    while d.weekday() != weekday:
        d += timedelta(days=1)
    while d.month == month:
        yield d
        d += timedelta(days=7)


class TestDateUtils(unittest.TestCase):
    """
    Test date utilities.
    """

    def test_get_next_market_day(self):
        """
        Test date_utils.get_next_market_day().
        """
        
        pairs = [
            ("2021-01-01", "2021-01-04"), # market holiday
            ("2025-01-06", "2025-01-07"), # market day
            ("2024-02-29", "2024-03-01"), # leap year
            ("2025-02-28", "2025-03-03"), # non-leap year, but Friday
            ("2024-11-01", "2024-11-04"), # DST transition
            ("2024-12-21", "2024-12-23"), # weekend
            ]
        with self.subTest('date only, no timezone'):
            for date, next_date in pairs:
                dt = datetime.strptime(date, "%Y-%m-%d")
                next_dt = datetime.strptime(next_date, "%Y-%m-%d")
                msg = f"date: {date}, next_date: {next_date}"
                self.assertEqual(get_next_market_day(dt).date(), next_dt.date(), msg)
                # without explicitly dropping the time
                self.assertEqual(get_next_market_day(dt), next_dt, msg)

        with self.subTest('entire datetime with timezone, midnight'):
            for date, next_date in pairs:
                dt = eastern.localize(datetime.strptime(date, "%Y-%m-%d"))
                next_dt = eastern.localize(datetime.strptime(next_date, "%Y-%m-%d"))
                msg = f"date: {date}, next_date: {next_date}"
                # timezone must be provided to ensure timing is correct on DST transition
                self.assertEqual(get_next_market_day(dt, eastern), next_dt, msg)

        # set the pairs to 08:00:00
        utc = timezone('UTC')
        with self.subTest('entire datetime with UTC timezone, 08:00'):
            for date, next_date in pairs:
                dt = utc.localize(datetime.strptime(date + " 08:00", "%Y-%m-%d %H:%M"))
                next_dt = utc.localize(datetime.strptime(next_date + " 08:00", "%Y-%m-%d %H:%M"))
                msg = f"date: {date}, next_date: {next_date}"
                self.assertEqual(get_next_market_day(dt), next_dt, msg)

        with self.subTest('additional dates, no timezone'):
            # wednesdays in March 2025
            for dt in day_generator(2025, 3, 'wed'):
                self.assertEqual(get_next_market_day(dt), dt + timedelta(1))

            # sundays in April 2025
            for dt in day_generator(2025, 4, 'sun'):
                self.assertEqual(get_next_market_day(dt), dt + timedelta(1))

            # saturdays in June 2025
            for dt in day_generator(2025, 6, 'sat'):
                self.assertEqual(get_next_market_day(dt), dt + timedelta(2))


    def test_get_prev_market_day(self):
        """
        Test date_utils.get_prev_market_day().
        """
        
        pairs = [
            ("2021-01-04", "2020-12-31"), # market holiday on previous Friday
            ("2025-01-07", "2025-01-06"), # market day
            ("2024-03-01", "2024-02-29"), # leap year
            ("2025-03-03", "2025-02-28"), # non-leap year, but Monday
            ("2024-11-04", "2024-11-01"), # DST transition
            ("2024-12-22", "2024-12-20"), # weekend
            ]
        with self.subTest('date only, no timezone'):
            for date, prev_date in pairs:
                dt = datetime.strptime(date, "%Y-%m-%d")
                prev_dt = datetime.strptime(prev_date, "%Y-%m-%d")
                msg = f"date: {date}, prev_date: {prev_date}"
                self.assertEqual(get_prev_market_day(dt).date(), prev_dt.date(), msg)
                # without explicitly dropping the time
                self.assertEqual(get_prev_market_day(dt), prev_dt, msg)
                
        with self.subTest('entire datetime with timezone, midnight'):
            for date, prev_date in pairs:
                dt = eastern.localize(datetime.strptime(date, "%Y-%m-%d"))
                prev_dt = eastern.localize(datetime.strptime(prev_date, "%Y-%m-%d"))
                # timezone must be provided to ensure timing is correct on DST transition
                self.assertEqual(get_prev_market_day(dt, eastern), prev_dt, msg)

        # set the pairs to 08:00:00
        utc = timezone('UTC')
        with self.subTest('entire datetime with UTC timezone, 08:00'):
            for date, prev_date in pairs:
                dt = utc.localize(datetime.strptime(date + " 08:00", "%Y-%m-%d %H:%M"))
                prev_dt = utc.localize(datetime.strptime(prev_date + " 08:00", "%Y-%m-%d %H:%M"))
                msg = f"date: {date}, prev_date: {prev_date}"
                self.assertEqual(get_prev_market_day(dt), prev_dt, msg)
        
        with self.subTest('additional dates, no timezone'):
            # wednesdays in March 2025
            for dt in day_generator(2025, 3, 'wed'):
                self.assertEqual(get_prev_market_day(dt), dt - timedelta(1))

            # saturdays in June 2025
            for dt in day_generator(2025, 6, 'sat'):
                self.assertEqual(get_prev_market_day(dt), dt - timedelta(1))

            # sundays in August 2025
            for dt in day_generator(2025, 8, 'sun'):
                self.assertEqual(get_prev_market_day(dt), dt - timedelta(2))


if __name__ == '__main__':
    unittest.main()



"""
Adherence to Data Privacy and Security Procedures
"""

P3_descriptors = [
    # field names from Apex SOD files
    'TaxIDNumber',
    'AccountNumber',
    # field names from roboadvisor backend
    'apex_account_id',
    'user_id' # cognito user id
]


# FUNCTIONS ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def check_recipients(recipients, read_authorization, classification= 'P2'):
    """
    Check that the recipients have P2 authorization.

    Parameters
    ----------
    recipients : str or list of str
        List of email addresses.
    read_authorization : dict
        Dictionary of email addresses and their privacy classification.
    classification : str, optional
        Privacy classification to check against. The default is 'P2'.

    Returns
    -------
    bool
        True if all recipients have the authorization, False otherwise.
    """
    if isinstance(recipients, str):
        recipients= [s.strip() for s in recipients.split(',')]

    for recipient in recipients:
        if recipient not in read_authorization.keys():
            return False
        else:
            if classification not in read_authorization[recipient]:
                return False
    return True

def below_size_limit(data):
    """
    Check that the data is below the size limit for P2 data. P2 data with more than 500 rows is not allowed.

    Parameters
    ----------
    data : pd.DataFrame or iterable
        Data to check.

    Returns
    -------
    bool
        True if the data is below the size limit, False otherwise.
    """
    # raise an error if the data is a string
    if isinstance(data, str):  raise TypeError('Data cannot be a string')
    
    P2_max_size = 500
    if len(data) > P2_max_size:
        return False
    else:
        return True
    
def contains_P3(content, exceptions=[]):
    """
    Check if the content contains P3 data.

    Parameters
    ----------
    content : str
        Email content.
    exceptions : list of str, optional
        List of P3 descriptors to ignore if they have been anonymized or replaced
        with non-P3 data. The default is [].

    Returns
    -------
    bool
        True if the content contains P3 data, False otherwise.
    """
    for descriptor in P3_descriptors:
        if (descriptor in content) and (descriptor not in exceptions):
            return True
    return False



"""
Email utility functions.
Copyright (C) 2022 Humankind Investments
"""

import os, re
import traceback
import numpy as np

from datetime import date
from dateutil.parser import parse

import smtplib
import imaplib
from email import message_from_bytes
from email.header import decode_header
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication


def send_email(content, subject, receiver, credentials, importance_level=None):
    """
    Send an email alert.

    Parameters
    ----------
    content : str
        Content of the email to send.
    subject : str
        Subject of the email to send.
    receiver : str
        Email to send the alert to.
    credentials : dict
        Provides the credentials for sender, password, and port.
    importance_level : str, optional
        The importance flag of the email. The default is None.
    """
    
    for key in ['sender', 'password', 'port']:
        assert key in credentials, f"credentials must have a key for {key}"
    
    sender = credentials['sender']
    passw = credentials['password']
    port = credentials['port']
    
    msg = MIMEText(content)
    msg['From'] = sender
    msg['To'] = receiver
    msg['Subject'] = subject
    msg.add_header('Content-Type', 'text/html')

    if importance_level: msg.add_header('Importance', importance_level)

    server = smtplib.SMTP('smtp.' + sender.split('@')[1], port)
    server.ehlo()
    server.starttls()
    server.ehlo()

    server.login(sender, passw)
    server.send_message(msg)
    server.quit()


def wrotetoday(platform, receivers, content, credentials):
    """
    Send an email notification that a script succeeded.

    Parameters
    ----------
    platform : str
        Name of the platform.
    receivers : list of str
        List of emails to send the alert to.
    content : str
        Content of the email alert.
    credentials : dict
        Provides the credentials for sender, password, and port.
    """
    
    subject = '[%s] finished writing %s' % (platform, date.today())

    if isinstance(receivers, list):
        receivers= ", ".join(receivers)
        send_email(content, subject, receivers, credentials)
    elif isinstance(receivers, str):
        send_email(content, subject, receivers, credentials)
    else:
        raise TypeError("receivers must be a str or list")

            
def error_alert(platform, receivers, filename, credentials, e=None):
    """
    Send an email alert that a script failed.

    Parameters
    ----------
    platform : str
        Name of the platform to be included in the subject line.
    receivers : list of str
        List of emails to send the alert to.
    filename : str
        Name of the script which ran into the error.
    credentials : dict
        Provides the credentials for sender, password, and port.
    e : Exception
        The exception that was raised.
    """
    
    subject = '[ERROR] %s failed %s' % (platform, date.today())
    if e is None:
        error = """No error message was provided. Include Exception in the 
        call to error_alert to include the error message."""
    else:
        error = ''.join(traceback.format_exception(None, e, e.__traceback__)).replace('    ', '&emsp;').replace(
                                '  ', '&ensp;').replace('\n', '<br>')
    content = f"""{filename} ran into the following error: <br> 
                  <span style="font-family: Courier;">{error}</span><br>
                  Check log for further details.
               """
    importance_level = 'High'
    if isinstance(receivers, list):
        receivers= ", ".join(receivers)
    elif isinstance(receivers, str):
        pass
    else:
        raise TypeError("receivers must be a str or list")
    
    send_email(content, subject, receivers, credentials, importance_level)


def send_email_with_attachment(content, subject, receiver, credentials, attachment_paths):
    """
    Send an email alert.

    Parameters
    ----------
    content : str
        Content of the email to send.
    subject : str
        Subject of the email to send.
    receiver : str
        Email to send the alert to.
    credentials : dict
        Provides the credentials for sender, password, and port.
    attachment_paths : list
        List of file path(s) for the attachment(s)
    """
    
    for key in ['sender', 'password', 'port']:
        assert key in credentials, f"credentials must have a key for {key}"
    
    if not isinstance(attachment_paths, list):
        attachment_paths= [attachment_paths]
        
    sender = credentials['sender']
    passw = credentials['password']
    port = credentials['port']
    
    # Create a multipart message
    msg = MIMEMultipart()
    msg['From'] = sender
    msg['To'] = receiver
    msg['Subject'] = subject
    msg.attach(MIMEText(content, _subtype='html'))
    msg.add_header('Content-Type', 'text/html')
        
    for path in attachment_paths:
        # Get the file name
        attachment_name = os.path.basename(path)
    
        # Read the attachment file
        with open(path, "rb") as file:
            attachment_part = MIMEApplication(
                file.read(),
                Name=attachment_name
            )
        attachment_part.add_header(
            "Content-Disposition",
            f"attachment; filename= {attachment_name}",
        )
    
        # Add the attachment to the message
        msg.attach(attachment_part)

    with smtplib.SMTP('smtp.' + sender.split('@')[1], port) as server:
        server.starttls()
        server.login(sender, passw)
        server.send_message(msg)


def download_attachments(response, folder_name, get_attachment=True):
    """
    Download the attachments from an email and save it to the given folder.

    Parameters
    ----------
    response : tuple
        Component of an email message, i.e. message envelope and data.
    folder_name : str
        The folder path in which to save the attachments.
    get_attachment : bool
        Get attachments if they exist. Otherwise, download email body.

    Returns
    -------
    str
        The names to which to save the files, excluding the path.
    """
    
    filenames = []
    if isinstance(response, tuple):
        # parse a bytes email into a message object
        msg = message_from_bytes(response[1])
        
        # decode the email subject
        subject, encoding = decode_header(msg["Subject"])[0]
        if isinstance(subject, bytes): subject = subject.decode(encoding)  # decode bytes to str
        subject = re.sub(r':', '', ' '.join(subject.split()))

        # if the email message is multipart
        if msg.is_multipart():
            # iterate over email parts
            for part in msg.walk():
                # download and save attachment if it exists
                content_disposition = str(part.get('Content-Disposition'))
                if 'attachment' in content_disposition and get_attachment:
                    filename = part.get_filename()
                    if filename:
                        filepath = os.path.join(folder_name, filename)
                        if not os.path.exists(folder_name): os.makedirs(folder_name)
                        with open(filepath, "wb") as f: f.write(part.get_payload(decode=True))
                        filenames.append(filename)
            # if no attachment found, iterate over email parts again and get plain text
            if not filenames:
                for part in msg.walk():
                    content_type = part.get_content_type()
                    if content_type == 'text/plain':
                        filename = subject + '.txt'
                        filepath = os.path.join(folder_name, filename)
                        if not os.path.exists(folder_name): os.makedirs(folder_name)
                        with open(filepath, 'wb') as f: f.write(part.get_payload(decode=True))
                        filenames.append(filename)
        else:
            # extract content type and body of email
            content_type = msg.get_content_type()
            body = msg.get_payload(decode=True).decode()

            if content_type == 'text/html' or content_type == 'text/plain':
                if content_type == 'text/html': ext = '.html'
                elif content_type == 'text/plain': ext = '.txt'
                filename = subject + ext
                filepath = os.path.join(folder_name, filename)
                if not os.path.exists(folder_name): os.makedirs(folder_name)
                with open(filepath, "w") as f: f.write(body)
                filenames.append(filename)

    return filenames


def readmails(username, password, tag, folder_name, sender,
              n_mails=np.inf, min_date='', get_attachment=True):
    """
    Download the attachments from the most recent emails with the desired tag.

    Parameters
    ----------
    username : str
        Email address from which to read messages.
    password : str
        Email password.
    tag : str
        A tag to identify messages in the inbox.
    folder_name : str
        The path of the folder to which to save the attachment.
    sender : str
        The email address of the desired sender. Emails from other senders will
        be excluded from the potential emails to read.
    n_mails : int
        The number of recent emails to read. Default reads all emails.
    min_date : str
        The minimum date on or after which to check for emails. Default checks all dates.
    get_attachment : bool
        Get attachments if they exist. Otherwise, download email body.

    Returns
    -------
    list
        List of file names, excluding the path.
    """

    # create an IMAP4 class with SSL
    imap = imaplib.IMAP4_SSL("imap.gmail.com")
    imap.login(username, password)

    # get total number of messages
    status, messages = imap.select(tag)
    messages = int(messages[0])

    # loop over relevant messages and select emails of interest
    datetimeorder = {}
    for p in range(messages, 0, -1):
        res, msg = imap.fetch(str(p), "(RFC822)")
        if res != 'OK': continue
        email_message = message_from_bytes(msg[0][1])
        email_from = email_message['From'].lower()
        pat = r'[^\s<]+@[^\s>]+'
        if re.findall(pat, email_from)[0] != sender.lower(): continue
        datetime_obj = parse(email_message['Date'])
        datetimeorder.update({(email_message['Subject'], datetime_obj) : p})
        if len(datetimeorder) >= n_mails or str(datetime_obj.date()) < min_date: break
    
    # get recent emails and download attachments
    filenamelst = []
    for index in datetimeorder.values():
        res, msg = imap.fetch(str(index), '(RFC822)')
        for response in msg:
            filenames = download_attachments(response, folder_name, get_attachment)
            if filenames:
                for filename in filenames: filenamelst.append(filename)
    
    return filenamelst




import numpy as np

from datetime import datetime, timedelta
import pytz
eastern = pytz.timezone('US/Eastern')
import pandas_market_calendars as mcal


def get_prev_market_day(dt, tz=None):
    """
    Get previous market day, accounting for weekends and market holidays.

    Parameters
    ----------
    dt : datetime.datetime
        Current datetime.

    Returns
    -------
    datetime.datetime
        Previous market day datetime.
    """
    if tz is not None: dt = dt.replace(tzinfo=None)

    ydt = dt - timedelta(1)  # yesterday

    # check if previous day the weekend
    if ydt.weekday() == 5: ydt -= timedelta(1)
    elif ydt.weekday() == 6: ydt -= timedelta(2)

    # check for holiday
    nyse_holidays = mcal.get_calendar('NYSE').holidays().holidays
    if ydt.date() in nyse_holidays: ydt -= timedelta(1)

    # check for weekend again
    if ydt.weekday() == 6: ydt -= timedelta(2)

    if tz is not None: ydt = tz.localize(ydt)
    
    return ydt


def get_prev_market_days(dt, n, ascending=True):
    """
    Get the previous n market days.

    Parameters
    ----------
    dt : datetime.datetime
        Date for which to get previous market days.
    n : int
        Number of previous market days to get.
    ascending : bool, optional
        Flag to sort dates in ascending (chronological) order. The default is True.

    Returns
    -------
    list of datetime.datetime
        List of previous market days.
    """

    prev_days = []
    while len(prev_days) < n:
        dt = get_prev_market_day(dt)
        prev_days.append(dt)
    
    if ascending: 
        return sorted(prev_days)
    else:
        return prev_days


def get_next_market_day(dt, tz=None):
    """
    Get next market day, accounting for weekends and market holidays.

    Parameters
    ----------
    dt : datetime.datetime
        Current datetime.
    tz : pytz.timezone, optional
        Timezone to use for the datetime. The hour in the output may 
        not match expectations unless a timezone is provided.

    Returns
    -------
    datetime.datetime
        Next market day datetime.
    """
    if tz is not None: dt = dt.replace(tzinfo=None)

    next_dt = dt + timedelta(1)

    # check if next day the weekend
    if next_dt.weekday() == 5: next_dt += timedelta(2)
    elif next_dt.weekday() == 6: next_dt += timedelta(1)

    # check for holiday
    nyse_holidays = mcal.get_calendar('NYSE').holidays().holidays
    if next_dt.date() in nyse_holidays: next_dt += timedelta(1)

    # check for weekend again
    if next_dt.weekday() == 5: next_dt += timedelta(2)

    if tz is not None: next_dt = tz.localize(next_dt)

    return next_dt


def get_quarter_start(dt):
    """
    Get start date of the current quarter.

    Parameters
    ----------
    dt : datetime.datetime
        Current datetime

    Returns
    -------
    datetime.datetime
        Datetime of last quarter end.
    """

    # get current month and year
    month = dt.month
    year = dt.year

    # calculate first day of current quarter
    quarter_start = datetime(year, ((month-1) // 3) * 3 + 1, 1).astimezone(eastern)
    return quarter_start


def get_last_quarter_end(dt):
    """
    Get date of last quarter end.

    Parameters
    ----------
    dt : datetime.datetime
        Current datetime

    Returns
    -------
    datetime.datetime
        Datetime of last quarter end.
    """

    # calculate first day of current quarter
    quarter_start = get_quarter_start(dt)

    # calculate last day of previous quarter
    last_quarter_end = eastern.normalize(quarter_start - timedelta(1))

    return last_quarter_end


def is_not_market_day(dt, nyse_holidays=None):
    """
    Check if a given date is a market holiday or weekend.

    Parameters
    ----------
    dt : datetime.datetime
        Date to check.
    nyse_holidays : tuple
        NYSE market holidays.

    Returns
    -------
    bool
        True if date is not a market day, False otherwise.
    """
    
    # check for weekend
    if dt.weekday() in [5, 6]: return True
    
    # check for holiday
    if nyse_holidays is None: nyse_holidays = mcal.get_calendar('NYSE').holidays().holidays
    if dt.strftime('%Y-%m-%d') in np.datetime_as_string(nyse_holidays): return True

    return False


def market_dates(start_date, end_date):
    """
    Get market dates between start and end dates.

    Parameters
    ----------
    start_date : datetime.datetime or str
        Start date. If str, must be in YYYY-MM-DD format.
    end_date : datetime.datetime or str
        End date. Not included in the output.
        If str, must be in YYYY-MM-DD format.

    Returns
    -------
    list of datetime.datetime
        List of market dates between start and end dates.
    """
    nyse_holidays = mcal.get_calendar('NYSE').holidays().holidays

    # if input dates are strings, convert to datetime
    if isinstance(start_date, str):
        start_date = datetime.strptime(start_date, "%Y-%m-%d")

    if isinstance(end_date, str):
        end_date = datetime.strptime(end_date, "%Y-%m-%d")

    dates = []
    while start_date < end_date:
        if (not start_date.date() in nyse_holidays) and (start_date.weekday() < 5):
            dates.append(start_date)
        start_date += timedelta(days=1)

    return dates
